//Zegar martycowy LED by Konrad Bekasiński
//Wersja 1.0.2
//26.05.2020

#include <ESP8266WiFi.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <ArduinoJson.h>
#include <ESP8266HTTPClient.h>

// -- Dane do połączenia się z siecią Wi-Fi -----------------------------------------------------------
#ifndef STASSID
#define STASSID "Nazwa-Sieci-Wi-Fi"
#define STAPSK  "Haslo-do-Wi-Fi"
#endif
const char* ssid = STASSID;
const char* password = STAPSK;


//ticker - pozwala na uruchamianie funkcji co określony czas, w tym przypadku co sekundę
#include <Ticker.h>
Ticker sss;
Ticker pomiary;
Ticker aktualizacja;


// -- biblioteki odnośnie obsłuki RTC
#include <Wire.h>
#include <DS3231.h>
DS3231 ds;
RTCDateTime dt;

// Biblioteki odnośnie czujnika BMP280
#include <Adafruit_BMP280.h>
Adafruit_BMP280 bmp;

//Biblioteki odnośnie czujnika DHT22
#include "DHT.h"
#define DHTPIN D3
#define DHTTYPE DHT22
DHT dht(DHTPIN, DHTTYPE);

//Biblioteki do obsługi wyświetlacza
#include <MD_Parola.h>
#include <MD_MAX72xx.h>
#include <SPI.h>
#define HARDWARE_TYPE MD_MAX72XX::FC16_HW
#define MAX_DEVICES 4
#define CS_PIN    D6
//DIN - D7; CLK - D5; CS - D6
//FC16_HW

MD_Parola P = MD_Parola(HARDWARE_TYPE, CS_PIN, MAX_DEVICES);
#define  BUF_SIZE  75
char curMessage[BUF_SIZE] = { "" }; //Zawartość zmiennej jest aktualnie wyświetlana na wyświetlaczu
char newMessage[BUF_SIZE] = { "" }; //Zmienna do przygotowania nowej wartości do wyświetlenia
bool newMessageAvailable = true;
bool animacja = true;

uint8_t scrollSpeed = 25; //szybkość przewijania się animacji
textEffect_t scrollEffect = PA_SCROLL_LEFT; //efekt animacji
textPosition_t scrollAlign = PA_CENTER; //wyrównanie tekstu
uint16_t scrollPause = 5000; //czas pauzy po wykonaniu animacji

//Startowe zmienne globalne programu
byte pozycja = 0; // określa czy wyświetlamy czas z dwukropkiem (1) czy bez (0)
word zmiana = 0; // określa czy wyświetlamy czas (0) czy inne parametry (1)
word zmiana2 = 1; // jeśli wyświetlamy coś innego niż czas, określa czy jest to temperatura(1), wilgotność(2), czy ciśnienie(3)
int licznik = 0; // zmienna do odliczania czasu pomiędzy aktualizacjami czasu
word godziny; //godziny
word minuty; //minuty
float wilg = 0; //wilgotność
float temp2 = 0; //
long rssi = 0; //siła sygnału WiFi
float temp_dht, temp_bmp, temp_rtc; //zmienne przechowujące temperaturę, odczytane kolejno z: DHT22, BMP280 i DS3231 - on też ma wbudowany termometr
int cis, sensorValue; //zmienna przechowująca ciśnienie oraz wartość odczytaną z ADC
String url = ""; // zmianna do przechowywania adresu URL, który jest używany przy pobieraniu czasu z serwera i wysyłaniu pomiarów na serwer
bool czyPomiar = false;
bool czyAktualizacja = false;

//Ustawienia -------------------------------------------------------------------------------------------------------------------------
#define CZAS 30 //przez ile sekund ma być wyświeltany czas
#define INNE 3 //przez ile sekuund mają być wyświetlane inne parametry
#define AKTUALIZACJA 3600 // co ile sekund ma być aktualizowany czas (w sekundach, 3600s = 1 godzina)
#define POMIARY 300 //co ile sekund ma być wykonywany pomiar do wysyłki na serwer (w sekundach, 300s = 5 min)



// DEBUGOWANIE
//Po przełączeniu na 1, program wysyła dane diagnostyczne przez port szeregowy.
//Jest to pomocne przy sprawdzaniu działania zegara i lokalizowaniu ewentualnych problemów
#define DEBUGOWANIE = 1


//--------------------------------------------------------setup()------------------------------------------------------------------
void setup(void)
{
  Serial.begin(9600);
#ifdef DEBUGOWANIE
  Serial.println("Startuje...");
  Serial.print("SSID: ");
  Serial.println(ssid);
  Serial.print("Haslo: ");
  Serial.println(password);
#endif

  //Opóźnienie okazało się niezbędne aby wyświetlacz wystartował prawidłowo
  delay(1500);

#ifdef DEBUGOWANIE
  Serial.println("Przed wyswietlaczem");
#endif
  //Startowanie wyświetlacza
  P.begin();
  P.print("ZEGAR");

  //Połączenie się z WiFi
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Serial.println("Polaczanie nieudane! Restart...");
    delay(5000);
    ESP.restart();
  }

#ifdef DEBUGOWANIE
  Serial.println("Po laczeniu...");
  Serial.println(WiFi.localIP());
#endif


  //Ustawienia odnośnie aktualizacji programu przez OTA - urządzenie będzie zgłaszać się jako "Zegar matrycowy"
  ArduinoOTA.setHostname("Zegar matrycowy - Płock");
 // ArduinoOTA.setHostname("Zegar matrycowy - drugi");
  ArduinoOTA.onStart([]() {
    String type;
    if (ArduinoOTA.getCommand() == U_FLASH) {
      type = "sketch";
    } else { // U_SPIFFS
      type = "filesystem";
    }
  });
  ArduinoOTA.begin();


  P.displayText(curMessage, scrollAlign, scrollSpeed, scrollPause, scrollEffect, scrollEffect);

  //Startowanie bibliotek peryferiów
  ds.begin();
  aktualizujCzas();
  bmp.begin(0x76);
  dht.begin();

#ifdef DEBUGOWANIE
  Serial.print("Adres IP: ");
  Serial.println(WiFi.localIP());
#endif

  //Pierwsze po uruchomieniu wykonanie pomiarów oraz wysłanie ich na serwer
  zrobPomiary();
  wyslijPomiary();


  //aktywowanie tickera aby odliczał czas co sekundę i uruchamiał funkcję wskazane funkcje
  sss.attach(1, wyswietl);
  pomiary.attach(POMIARY, setPomiar);
  aktualizacja.attach(AKTUALIZACJA, setAkualizacja);
}

//-------------------------------------------------------------loop()-------------------------------------------------------------------------------
void loop(void)
{
  //Niezbędne do obsługi programowania przez WiFi
  ArduinoOTA.handle();

  //Zmiana wyświetlania na temperaturę, wilgotność i ciśnienie jest animowana, gdy wyświetlany jest czas nie ma animacji.
  //Poniższy kod jest potrzebny, gdy używamy animacji, zatem najpierw sprawdzam czy animacjia ma być w ogóle pokazana.
  if (animacja)
  {
    if (P.displayAnimate())
    {
      //Jeśli przygotowano nowy tekst do wyświetlenia
      if (newMessageAvailable)
      {
        strcpy(curMessage, newMessage);
        newMessageAvailable = false;
      }
      P.displayReset();
    }
  }

  //Sprawdzenia flagi, która oznacza, że trzeba wykonac pomiar a nastęonie jej zerowanie
  if (czyPomiar)
  {
    zrobPomiary();
    wyslijPomiary();
    czyPomiar = false;
  }

  //Sprawdzenia flagi, która oznacza, że trzeba zaktualizować czas a następnie jej zerowanie
  if (czyAktualizacja)
  {
    aktualizujCzas();
    czyAktualizacja = false;
  }

}

//-----------------------------------------------------------wyswietl()----------------------------------------------------------------------------
//Najważniejsza funkcja programu - wykonuje się co sekundę i wzależności od stanu zmiennych zmiana i zmiana2 wyświetla albo czas
//albo temperaturę, wilgotność i ciśnienie. Dodatkowo co sekundę zmienia się format czasu - z dwukropkiem i bez niego
void wyswietl()
{
  // Jeśli zmiana jest poniżej zawartości zdefiniowanej w CZAS (w sekundach) to wyświetlamy czas
  if (zmiana <= CZAS )
  {
    //pobranie czasu z RTC
    dt = ds.getDateTime();

    //Wyłączenie animacji zmiany zawartości ekranu
    animacja = false;
    newMessageAvailable = false;

    // z dwukropkiem
    if (pozycja == 1)
    {
      P.print(ds.dateFormat("H:i", dt));
      // bez dwukkropka
    } else
    {
      P.print(ds.dateFormat("H i", dt));
    }
    // Jeśli zmiana jest powyżej wartości określonej dla czasu - wyświetlamy pomiary włączając animację
  } else if (zmiana == (CZAS + 1))
  {
    //Wykonanie pomiarów do wyświetlenia
    zrobPomiary();

    // Jeśli zmiana2 = 1 to wyświetlamy temperaturę
    if (zmiana2 == 1)
    {
      scrollEffect = PA_SCROLL_DOWN;

      Serial.print(temp_dht);
      char result[4] = {""};
      dtostrf(temp_dht, 3, 1, result);
      sprintf(newMessage, "%s%cC", result, 247);
    }
    // Jeśli zmiana2 = 2 to wyświetlamy wilgotnosc
    else if (zmiana2 == 2)
    {
      Serial.print("wilgotnosc: ");
      Serial.println(wilg);
      char result[4] = {""};
      dtostrf(wilg, 3, 1, result);
      sprintf(newMessage, "%s%%", result);
      scrollEffect = PA_SCROLL_LEFT;
    }
    // Jeśli zmiana2 = 2 to wyświetlamy cisnienie
    //    else if (zmiana2 == 3)
    //    {
    //      scrollEffect = PA_SCROLL_UP;
    //      cis = bmp.readPressure() / 100;
    //      sprintf(newMessage, "%d", cis);
    //    }


    strcpy(curMessage, newMessage);
    P.displayText(curMessage, scrollAlign, scrollSpeed, scrollPause, scrollEffect, scrollEffect);

    //Dajemy znać, by uruchomić animację
    animacja = true;
    //Dajemy znać, że jest nowa wartość do wyświetlenia
    newMessageAvailable = true;

    //inkrementacja zmiennej i w razie potrzeby resetowanie wartości
    zmiana2++;
    if (zmiana2 > 2) zmiana2 = 1;
  }

  //Aktualizacja czasu jeśli spełniony warunek długości interwału
  if (licznik >= AKTUALIZACJA) aktualizujCzas();

  //Ustawienie parametrów animacji
  P.setTextEffect(scrollEffect, scrollEffect);
  P.setPause(scrollPause);

  //Resetowanie zmiennej decydującej o wyświetlaniu dwukropka
  pozycja = !pozycja;

  //inkrementacja kolejnych zmiennych
  zmiana++;
  licznik++;

  //Jeśli minęło już tyle ile ma być wyświetlany czas + tyle ile ma być wyświetlany pomiat to resetujemy licznik
  if (zmiana > (CZAS + INNE)) zmiana = 0;




#ifdef DEBUGOWANIE
  Serial.println(newMessage);
  Serial.print("Licznik: ");
  Serial.println(licznik);
#endif

  //Ustawiania jasności ekranu
  ustawJasnosc();
}

//------------------------------------------------------------ustawJasnosc()------------------------------------------------------------------
//Funkcja ustawiania jasności wyświetlacza na podstawie oddczytu z ADC (dzielnik napięcia na fotorezystorze)
void ustawJasnosc()
{
  sensorValue = 0;
  int jasnosc = 0;

  //Odczyt z przetwornika analogowo-cyfrowego (ADC0)
  sensorValue = analogRead(A0);

  //Mapowanie wyniku z ADC (zakres 0-1023) na poziomy jasności (0-15)
  jasnosc = map(sensorValue, 0, 1023, 0, 15);

#ifdef DEBUGOWANIE
  Serial.print("ADC: ");
  Serial.print(sensorValue);
  Serial.print("  jasnosc: ");
  Serial.println(jasnosc);
#endif

  //Właściwe ustawienie jasności ekranu
  P.setIntensity(jasnosc);
}


// --------------------------------------------------------------aktualizujCzas()------------------------------------------------------------
// -- aktualizuje czas przez WiFi (pobierany JSON przygotowany przez skrypt PHP)
bool aktualizujCzas()
{
#ifdef DEBUGOWANIE
  Serial.println("Aktualizacja czasu...");
#endif

  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    http.begin("http://meteo.kbprogram.pl/czas.php");
    int httpCode = http.GET();

    if (httpCode > 0) {

      const size_t bufferSize = JSON_OBJECT_SIZE(10) + JSON_OBJECT_SIZE(10) + JSON_OBJECT_SIZE(10) + JSON_OBJECT_SIZE(8) + 370;
      DynamicJsonBuffer jsonBuffer(bufferSize);
      JsonObject& root = jsonBuffer.parseObject(http.getString());

      // zmienne pobrane z JSON
      int sekunda = root["seconds"];
      int minuta = root["minutes"];
      int godzina = root["hours"];
      int dzien = root["mday"];
      int miesiac = root["mon"];
      int rok = root["year"];

#ifdef DEBUGOWANIE
      Serial.print("Pobrany czas: ");
      Serial.print(dzien);
      Serial.print(".");
      Serial.print(miesiac);
      Serial.print(".");
      Serial.print(rok);
      Serial.print(" ");
      Serial.print(godzina);
      Serial.print(":");
      Serial.print(minuta);
      Serial.print(":");
      Serial.println(sekunda);
#endif


      //Wpisywanie aktualnego czasu do RTC
      ds.setDateTime(rok, miesiac, dzien, godzina, minuta, sekunda);
    }
    http.end();
    //Zerujemy licznik po wykonaniu aktualizacji
    licznik = 0;
    return true;
  }
}



// ----------------------------------------------------------------zrobPomiary()----------------------------------------------------------------
void zrobPomiary()
{
  //Pomiary z czujnika DHT - w pętli dopóki pomiar nie będzie prawidłowy - taki urok 1-wire
  do
  {
    wilg = dht.readHumidity();
    temp_dht = dht.readTemperature();
    delay(100);
  } while (isnan(wilg) or isnan(temp_dht));

  //Pomiary z czujnika BMP
  cis = bmp.readPressure() / 100;
  temp_bmp = bmp.readTemperature();

  //Pomiar z zegara RTC
  temp_rtc = ds.readTemperature();

  //Odczyt z ADC
  sensorValue = analogRead(A0);

#ifdef DEBUGOWANIE
  Serial.println("Wyniki pomiarow:");
  Serial.print("Temperatura:  DHT: ");
  Serial.print(temp_dht);
  Serial.print(", BMP: ");
  Serial.print(temp_bmp);
  Serial.print(". RTC: ");
  Serial.println(temp_rtc);
  Serial.print("Wilgotnosc: ");
  Serial.println(wilg);
  Serial.print("Cisnienie: ");
  Serial.println(cis);
  Serial.print("ADC: ");
  Serial.println(sensorValue);
#endif
}



//-----------------------------------------------------------------wyslijPomiary()-----------------------------------------------------------
bool wyslijPomiary()
{
  bool wynik = false;
#ifdef DEBUGOWANIE
  Serial.println("------------------------Wyslanie pomiarow na serwer------------------------");
#endif

  // Check WiFi Status
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    url = "http://meteo.kbprogram.pl/dodaj_pomiar.php?id_czujnika=5&temp1=";

    rssi = WiFi.RSSI();

    url += temp_dht;
    url += "&temp2=";
    url += temp_bmp;
    url += "&cisnienie=";
    url += cis;
    url += "&naslonecznienie=";
    url += sensorValue;
    url += "&wilgotnosc=";
    url += wilg;
    url += "&rssi=";
    url += rssi;

#ifdef DEBUGOWANIE
    Serial.println(url);
#endif

    http.begin(url);

    int httpCode = http.GET();


    if (httpCode > 0) {
#ifdef DEBUGOWANIE
      Serial.print("Odpowiedz HTTP: ");
      Serial.println(httpCode);
#endif

      const size_t bufferSize = JSON_OBJECT_SIZE(10) + JSON_OBJECT_SIZE(10) + JSON_OBJECT_SIZE(10) + JSON_OBJECT_SIZE(8) + 370;
      DynamicJsonBuffer jsonBuffer(bufferSize);
      JsonObject& root = jsonBuffer.parseObject(http.getString());

      int status_json = root["status"];


#ifdef DEBUGOWANIE
      Serial.print("Ststus z serwera w JSON: ");
      Serial.println(status_json);
#endif

      if (status_json == 1)
      {
        wynik = true;
      }
      else
      {
#ifdef DEBUGOWANIE
        Serial.println("Blad po stronie serwera");
#endif
        wynik = false;
      }

    } else {
      Serial.print("ERROR, HTTP code: ");
      Serial.println(httpCode);
    }
    Serial.println("HTTP end()");
    http.end();
  } else {
    Serial.println("Nie polaczono z WiFi");
  }
}

void setPomiar()
{
  czyPomiar = true;
}

void setAkualizacja()
{
  czyAktualizacja = true;
}
